/* $USAGI: ninfod_core.c,v 1.20 2002/12/09 06:45:12 yoshfuji Exp $ */
/*
 * Copyright (C) 2002 USAGI/WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 * Author:
 * 	YOSHIFUJI Hideaki <yoshfuji@linux-ipv6.org>
 */

#if HAVE_CONFIG_H
#include "config.h"
#endif

#if HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif
#if STDC_HEADERS
# include <stdio.h>
# include <stdlib.h>
# include <stddef.h>
#else
# if HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif
#if ENABLE_THREADS && HAVE_PTHREAD_H
# include <pthread.h>
#endif
#if HAVE_STRING_H
# if !STDC_HEADERS && HAVE_MEMORY_H
#  include <memory.h>
# endif
# include <string.h>
#endif
#if HAVE_STRINGS_H
# include <strings.h>
#endif
#if HAVE_INTTYPES_H
# include <inttypes.h>
#else
# if HAVE_STDINT_H
#  include <stdint.h>
# endif
#endif
#if HAVE_UNISTD_H
# include <unistd.h>
#endif

#if TIME_WITH_SYS_TIME   
# include <sys/time.h>  
# include <time.h>
#else
# if HAVE_SYS_TIME_H     
#  include <sys/time.h>
# else                
#  include <time.h>
# endif                  
#endif                   

#if HAVE_SYS_UIO_H
#include <sys/uio.h>
#endif

#if HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#if HAVE_NETINET_ICMP6_H
# include <netinet/icmp6.h>
#endif
#ifndef HAVE_STRUCT_ICMP6_NODEINFO
# include "icmp6_nodeinfo.h"
#endif

#if HAVE_NETDB_H
# include <netdb.h>
#endif
#include <errno.h>

#if HAVE_SYSLOG_H
# include <syslog.h>
#endif

#include "ninfod.h"

#ifndef offsetof
# define offsetof(aggregate,member)	((size_t)&((aggregate *)0)->member)
#endif

/* ---------- */
/* ID */
static char *RCSID __attribute__ ((unused)) = "$USAGI: ninfod_core.c,v 1.20 2002/12/09 06:45:12 yoshfuji Exp $";

/* Variables */
int initialized = 0;

#if ENABLE_THREADS && HAVE_LIBPTHREAD
pthread_attr_t pattr;
#endif

static uint32_t suptypes[(MAX_SUPTYPES+31)>>5];
static size_t suptypes_len;

/* ---------- */
struct subjinfo {
	uint8_t	code;
	char	*name;
	int	(*checksubj)(CHECKANDFILL_ARGS);
	int	(*init)(INIT_ARGS);
};

static struct subjinfo subjinfo_table [] = {
	[ICMP6_NI_SUBJ_IPV6] = {
		.code = ICMP6_NI_SUBJ_IPV6,
		.name = "IPv6",
		//.init = init_nodeinfo_ipv6addr,
		.checksubj = pr_nodeinfo_ipv6addr,
	},
	[ICMP6_NI_SUBJ_FQDN] = {
		.code = ICMP6_NI_SUBJ_FQDN,
		.name = "FQDN",
		//.init = init_nodeinfo_nodename,
		.checksubj = pr_nodeinfo_nodename,
	},
	[ICMP6_NI_SUBJ_IPV4] = {
		.code = ICMP6_NI_SUBJ_IPV4,
		.name = "IPv4",
		//.init = init_nodeinfo_ipv4addr,
		.checksubj = pr_nodeinfo_ipv4addr,
	},
};

static struct subjinfo subjinfo_null = {
	.name = "null",
	.checksubj = pr_nodeinfo_noop,
};

static __inline__ struct subjinfo *subjinfo_lookup(int code)
{
	if (code >= sizeof(subjinfo_table)/sizeof(subjinfo_table[0]))
		return NULL;
	if (subjinfo_table[code].name == NULL)
		return NULL;
	return &subjinfo_table[code];
}

/* ---------- */
#define QTYPEINFO_F_RATELIMIT	0x1

struct qtypeinfo {
	uint16_t qtype;
	char	*name;
	int	(*getreply)(CHECKANDFILL_ARGS);
	void	(*init)(INIT_ARGS);
	int	flags;
};

static struct qtypeinfo qtypeinfo_table[] = {
	[NI_QTYPE_NOOP]		= {
		.qtype = NI_QTYPE_NOOP,
		.name = "NOOP",
		.getreply = pr_nodeinfo_noop,
	},
	[NI_QTYPE_SUPTYPES]	= {
		.qtype = NI_QTYPE_SUPTYPES,
		.name = "SupTypes",
		.getreply = pr_nodeinfo_suptypes,
		.init = init_nodeinfo_suptypes,
	},
	[NI_QTYPE_DNSNAME]	= {
		.qtype = NI_QTYPE_DNSNAME,
		.name = "DnsName",
		.getreply = pr_nodeinfo_nodename,
		.init = init_nodeinfo_nodename,
	},
#if 1
	[NI_QTYPE_NODEADDR]	= {
		.qtype = NI_QTYPE_NODEADDR,
		.name = "NodeAddr",
		.getreply = pr_nodeinfo_ipv6addr,
		.init = init_nodeinfo_ipv6addr,
	},
#endif
#if 1
	[NI_QTYPE_IPV4ADDR]	= {
		.qtype = NI_QTYPE_IPV4ADDR,
		.name = "IPv4Addr",
		.getreply = pr_nodeinfo_ipv4addr,
		.init = init_nodeinfo_ipv4addr,
	},
#endif
};

static struct qtypeinfo qtypeinfo_unknown = {
	.name = "unknown",
	.getreply = pr_nodeinfo_unknown,
	.flags = 0,
};

static struct qtypeinfo qtypeinfo_refused = {
	.name = "refused",
	.getreply = pr_nodeinfo_refused,
	.flags = 0,
};

static __inline__ struct qtypeinfo *qtypeinfo_lookup(int qtype)
{
	if (qtype >= sizeof(qtypeinfo_table)/sizeof(qtypeinfo_table[0]))
		return &qtypeinfo_unknown;
	if (qtypeinfo_table[qtype].name == NULL)
		return &qtypeinfo_unknown;
	return &qtypeinfo_table[qtype];
}

/* ---------- */
/* noop */
int pr_nodeinfo_noop(CHECKANDFILL_ARGS)
{
	struct icmp6_nodeinfo *replybuf = NULL;

	DEBUG(LOG_DEBUG, "%s()\n", __FUNCTION__);

	if (subjlen) {
		DEBUG(LOG_WARNING,
		      "%s(): invalid subject length(%d)\n",
		      __FUNCTION__, subjlen);
		return 0;
	}

	if (reply) {
		replybuf = ni_malloc(sizeof(*replybuf));
		if (!replybuf)
			return -1;

		memset(replybuf, 0, sizeof(*replybuf));
		replybuf->ni_type = ICMP6_NI_REPLY;
		replybuf->ni_code = ICMP6_NI_SUCCESS;
		replybuf->ni_cksum = 0;
		replybuf->ni_qtype = htons(NI_QTYPE_NOOP);
		replybuf->ni_flags = flags;
		*reply = replybuf;
	}

	if (subj_if)
		*subj_if = 0;

	return (sizeof(replybuf));
}

/* suptypes */
int pr_nodeinfo_suptypes(CHECKANDFILL_ARGS)
{
	size_t replylen = sizeof(struct icmp6_nodeinfo) + (suptypes_len<<2);

	DEBUG(LOG_DEBUG, "%s()\n", __FUNCTION__);

	if (subjlen) {
		DEBUG(LOG_WARNING, "%s(): invalid subject length(%d)\n",
		      __FUNCTION__, subjlen);
		return 0;
	}

	if (reply) {
		struct icmp6_nodeinfo *replybuf = ni_malloc(replylen);
		if (!replybuf)
			return -1;

		memset(replybuf, 0, replylen);

		replybuf->ni_type = ICMP6_NI_REPLY;
		replybuf->ni_code = ICMP6_NI_SUCCESS;
		replybuf->ni_cksum = 0;
		replybuf->ni_qtype = htons(NI_QTYPE_SUPTYPES);
		replybuf->ni_flags = flags&~NI_SUPTYPE_FLAG_COMPRESS;
		
		memcpy((replybuf+1), suptypes, suptypes_len<<2);
		*reply = replybuf;
	}
	return replylen;
}

void init_nodeinfo_suptypes(INIT_ARGS)
{
	size_t w, b;
	int i;

	if (!forced && initialized)
		return;

	memset(suptypes, 0, sizeof(suptypes));
	suptypes_len = 0;

	for (i=0; i<sizeof(qtypeinfo_table)/sizeof(qtypeinfo_table[0]); i++) {
		unsigned short qtype;

		if (qtypeinfo_table[i].name == NULL)
			continue;
		qtype = qtypeinfo_table[i].qtype;
		w = qtype>>5;
		b = qtype&0x1f;
		if (w > sizeof(suptypes)/sizeof(suptypes[0])) {
			DEBUG(LOG_ERR, "Warning: Too Large Supported Types\n");
			exit(1);
		}
		suptypes[w] |= htonl(1<<b);

		if (suptypes_len < w)
			suptypes_len = w;
	}
	suptypes_len++;
}

/* ---------- */
/* unknown qtype response */
int pr_nodeinfo_unknown(CHECKANDFILL_ARGS)
{
	static struct icmp6_nodeinfo replybuf;

	if (!reply)
		return 0;

	replybuf.ni_type = ICMP6_NI_REPLY;
	replybuf.ni_code = ICMP6_NI_UNKNOWN;
	replybuf.ni_cksum = 0;
	//replybuf.ni_qtype = 0;
	replybuf.ni_flags = flags;

	*reply = &replybuf;
	return (sizeof(replybuf));
}

/* refused response */
int pr_nodeinfo_refused(CHECKANDFILL_ARGS)
{
	static struct icmp6_nodeinfo replybuf;

	if (!reply)
		return 0;

	replybuf.ni_type = ICMP6_NI_REPLY;
	replybuf.ni_code = ICMP6_NI_REFUSED;
	replybuf.ni_cksum = 0;
	//replybuf.ni_qtype = 0;
	replybuf.ni_flags = flags;

	*reply = &replybuf;
	return (sizeof(replybuf));
}

/* ---------- */
void init_core(int forced)
{
	int i;

	DEBUG(LOG_DEBUG, "%s()\n", __FUNCTION__);

	if (!initialized || forced) {
		struct timeval tv;
		unsigned int seed = 0;
		pid_t pid;

		if (gettimeofday(&tv, NULL) < 0) {
			DEBUG(LOG_WARNING, "%s(): failed to gettimeofday()\n", __FUNCTION__);
		} else {
			seed = (tv.tv_usec & 0xffffffff);
		}

		pid = getpid();
		seed ^= (((unsigned long)pid) & 0xffffffff);

		srand(seed);

#if ENABLE_THREADS && HAVE_LIBPTHREAD
		if (initialized)
			pthread_attr_destroy(&pattr);

		pthread_attr_init(&pattr);
		pthread_attr_setdetachstate(&pattr, PTHREAD_CREATE_DETACHED);
#endif
	}

	for (i=0; i<sizeof(subjinfo_table)/sizeof(subjinfo_table[0]); i++) {
		if (subjinfo_table[i].name == NULL)
			continue;
		if (subjinfo_table[i].init)
			subjinfo_table[i].init(forced);
	}

	for (i=0; i<sizeof(qtypeinfo_table)/sizeof(qtypeinfo_table[0]); i++) {
		if (qtypeinfo_table[i].name == NULL)
			continue;
		if (qtypeinfo_table[i].init)
			qtypeinfo_table[i].init(forced);
	}

	initialized = 1;

	return;
}

#if ENABLE_THREADS && HAVE_LIBPTHREAD
static void *ni_send_thread(void *data)
{
	ni_send(data);
	return NULL;
}
#else
static int ni_send_fork(struct packetcontext *p)
{
	pid_t child = fork();
	if (child < 0)
		return -1;
	if (child == 0) {
		pid_t grandchild = fork();
		if (grandchild == 0) {
			DEBUG(LOG_DEBUG, "%s(): worker=%d\n",
			      __FUNCTION__, getpid());
			ni_send(p);
		}
		exit(0);
	} else {
		waitpid(child, NULL, 0);
	}
	return 0;
}
#endif

int pr_nodeinfo(struct packetcontext *p)
{
	struct icmp6_nodeinfo *query = (struct icmp6_nodeinfo *)p->query;

	char *subject = (char *)(query + 1);
	size_t subjlen;
	struct subjinfo *subjinfo;
	struct qtypeinfo *qtypeinfo;
	int replyonsubjcheck = 0;
	unsigned int subj_if;
#if ENABLE_DEBUG
	char printbuf[128];
	int i;
	char *cp;
#endif
#if ENABLE_THREADS && HAVE_PTHREAD_H
	pthread_t thread;
#endif

	/* Step 1: Check length */
	if (p->querylen < sizeof(struct icmp6_nodeinfo)) {
		DEBUG(LOG_WARNING, "Query too short\n");
		ni_free(p);
		return -1;
	}

#if ENABLE_DEBUG
	cp = printbuf;
	for (i = 0; i < sizeof(query->ni_nonce); i++) {
		cp += sprintf(cp, " %02x", query->ni_nonce[i]);
	}
	DEBUG(LOG_DEBUG, "%s(): qtype=%d, flags=0x%04x, nonce[] = {%s }\n",
	      __FUNCTION__,
	      ntohs(query->ni_qtype), ntohs(query->ni_flags), printbuf);
#endif

	subjlen = p->querylen - sizeof(struct icmp6_nodeinfo);

	/* Step 2: Check Subject Code */
	switch(htons(query->ni_qtype)) {
	case NI_QTYPE_NOOP:
	case NI_QTYPE_SUPTYPES:
		if (query->ni_code != ICMP6_NI_SUBJ_FQDN) {
			DEBUG(LOG_WARNING,
			      "%s(): invalid/unknown code %u\n",
			      __FUNCTION__, query->ni_code);
			subjlen = 0;
		}
		subjinfo = &subjinfo_null;
		break;
	default:
		subjinfo = subjinfo_lookup(query->ni_code);
		if (!subjinfo) {
			DEBUG(LOG_WARNING,
			      "%s(): unknown code %u\n",
			      __FUNCTION__, query->ni_code);
			ni_free(p);
			return -1;
		}
	}

	/* Step 3: Lookup Qtype */
	qtypeinfo = qtypeinfo_lookup(ntohs(query->ni_qtype));

	/* Step 4: Check Subject
	 *         (And fill reply if it is available now)
	 */
	if (qtypeinfo->getreply == subjinfo->checksubj)
		replyonsubjcheck = 1;

	p->replylen = subjinfo->checksubj(subject, subjlen,
					  query->ni_flags,
					  &p->pktinfo,
					  replyonsubjcheck ? NULL : &subj_if,
					  replyonsubjcheck ? &p->reply : NULL);
	if (!p->replylen || (replyonsubjcheck && !p->reply)) {
		ni_free(p);
		return -1;
	}
	if (p->replylen < 0) {
		DEBUG(LOG_WARNING, 
		      "failed to make reply: %s\n", 
		      strerror(errno));
		ni_free(p);
		return -1;
	}

	/* XXX: Step 5: Check the policy */
	if (0) {
		if (p->reply) {
			ni_free(p->reply);
			p->reply = NULL;
		}
		replyonsubjcheck = 0;
		qtypeinfo = &qtypeinfo_refused;
	}

	/* Step 6: Fill the reply if not yet done */
	if (!replyonsubjcheck) {
		p->replylen = qtypeinfo->getreply(NULL, 0,
						  query->ni_flags,
						  &p->pktinfo,
						  &subj_if,
						  &p->reply);
		if (!p->replylen || !p->reply) {
			ni_free(p);
			return -1;
		}
		if (p->replylen < 0) {
			DEBUG(LOG_WARNING,
			      "failed to make reply: %s\n",
			      strerror(errno));
			ni_free(p);
			return -1;
		}
	}

	/* XXX: Step 7: Rate Limit */
	if (qtypeinfo->flags&QTYPEINFO_F_RATELIMIT && 0) {
		if (p->reply) {
			ni_free(p->reply);
			p->reply = NULL;
		}
		ni_free(p);
		return -1;
	}

	/* Step 8: Fill Qtype / Nonce */
	p->reply->ni_qtype = query->ni_qtype;
	memcpy(p->reply->ni_nonce, query->ni_nonce, sizeof(p->reply->ni_nonce));

	/* Step 9: Source address selection */
	if (IN6_IS_ADDR_MULTICAST(&p->pktinfo.ipi6_addr)) {
		/* if query was sent to multicast address,
		 * use source address selection in kernel.
		 * XXX: anycast?
		 */
		memset(&p->pktinfo.ipi6_addr, 0, sizeof(p->pktinfo.ipi6_addr));

	 	/* Random Delay between zero and MAX_ANYCAST_DELAY_TIME is
		 * required if query was sent to anycast or multicast address.
		 */
		p->delay = (int) (MAX_ANYCAST_DELAY_TIME*rand()/(RAND_MAX+1.0));
	} else {
		p->delay = 0;
	}

	/* Step 10: Send the reply
	 * XXX: with possible random delay */
#if ENABLE_THREADS && HAVE_LIBPTHREAD
	pthread_create(&thread, &pattr, ni_send_thread, p);	/*frees p*/
#else
	ni_send_fork(p);	/* frees p */
#endif

	return 0;
}

